/*
** Module   : RadioButtonImpl.java
** Abstract :
**
** Copyright (c) 2011-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 SIY 20111129 Extracted CHUI-specific code from RadioButton
** 002 SIY 20111221 Fixed constructor visibility.
** 003 GES 20140627 Changed color attribute method names to match naming standards.
** 004 CA  20140805 Do not allow ChUI impl classes access to APIs outside of ChuiOutputManager.
** 005 HC  20140911 GUI coordinate support - data types refactoring.
** 006 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 007 IAS 20150216 Added support for the EXPAND attribute and fix drawing
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.util.*;

/**
 * CHUI-specific portion of RadioButton.
 */
@SuppressWarnings("unchecked")
public class RadioButtonImpl
extends RadioButton<ChuiOutputManager>
{
   /**
    * Use this constructor when you want to set both the label and the value.
    * 
    * @param   config
    *          Reference to config.
    * @param   label
    *          The button's label which can clarify the button purpose.
    * @param   value
    *          The data which is the button's value.
    */
   RadioButtonImpl(RadioSetConfig config, String label, BaseDataType value)
   {
      super(config, label, value);
   }

   /**
    * Called by this RadioButton's parent container.
    */
   @Override
   public void draw()
   {
      Color dc = UiUtils.getDColor(this);
      Color pf = config().pfcolor;

      Point p = screenLocation();
      screen().at(p);

      if (autoResize)
      {
         drawButton(p, isEnabled() && parent().isEnabled() && focused, pf, dc);

         draw(p, getText(), dc);
      }
      else
      {
         // in chui char coordinates are always whole numbers
         int wd = (int) width - 3;
         if (wd >= 0)
         {
            drawButton(p, isEnabled() && focused, pf, dc);
         }

         if (wd > 0)
         {
            String ct = getText();
            draw(p, wd > ct.length() ?
                  String.format("%1$-" + wd + "s", ct) : ct.substring(0, wd),
                  dc);
         }
      }
   }

   /**
    * Draw cursor at the correct position.
    * 
    * @see com.goldencode.p2j.ui.client.widget.Widget#drawCaret()
    */
   @Override
   public void drawCaret()
   {
      if (!isFocused() || !isEnabled())
         return;
      
      screen().at(screenLocation().offset(1, 0));      
   }

   /**
    * Draw characters and advance position
    * @param p
    *       Starting point
    * @param cs
    *       The character sequence
    * @param dc
    *       The color
    */
   private void draw(Point p, CharSequence cs, Color dc)
   {
      for (int i = 0; i < cs.length(); i++)
      {
         screen().at(p).append(cs.charAt(i), dc);
         p.x++;
         
      }
   }

   /**
    * Draws radio button ("(X)" or "( )") and advance position
    *
    * @param p
    *       Starting point
    * @param highlighted
    *        Indicates whether the button is highlighted.
    * @param pf
    *        The color used inside this radio button.
    * @param dc
    *        The color which corresponds DCOLOR attribute for this radio
    *        button.
    */
   private void drawButton(Point p, boolean highlighted, Color pf, Color dc)
   {
      draw(p, isSelected() ? "(X)" : "( )", highlighted ? pf : dc);
   }
}
