/*
** Module   : handle.java
** Abstract : Progress 4GL compatible handle object
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- -----------------------Description-----------------------------------
** 001 GES 20051125   @23505 Created initial version which represents
**                           the Progress handle data type.
** 002 GES 20060206   @24300 Made compareTo() safe to process any object
**                           type (avoids ClassCastException).
** 003 GES 20060208   @24390 Added a default format string getter.
** 004 GES 20060724   @28161 Added a forced assign() abstract method.
** 005 GES 20060807   @28479 Moved to externalizable interface to optimize
**                           network performance. Throw an exception if
**                           an attempt is made to serialize this class
**                           since it makes no sense to transfer this
**                           between systems.
** 006 GES 20070219   @32175 Added support for the widget-handle built-in
**                           function.
** 007 GES 20070221   @32195 Added support for the TransparentWrapper as
**                           a contained object that exposes its
**                           referenced object instead.
** 008 GES 20080402   @37825 Added a new assign(Object) signature and a
**                           new constructor handle(WrappedResource).
** 009 CA  20090227   @41383 Undefined SELF or FOCUS handles refers a
**                           NullWidget instance - modified the isUnknown
**                           method to return true when the referenced
**                           value is not null, but unknown.
** 010 ECF 20101003          Removed ContextContainer inner class; replaced with
**                           parameterized use of ContextLocal. Fixed javadoc.
** 011 CA  20130112          Major rework to implement generic handles. 
**                           Added c'tor which builds a new instance from a
**                           BaseDataType value, if possible. This should be
**                           used only when constructing an instance using the
**                           value returned by a DYNAMIC-FUNCTION call; 4GL
**                           does some special casting in this case, which is
**                           not implemented yet.
** 012 EVL 20130130          Adding wrap methods to invoke DOM XML object
**                           method calls. The new supported class: XEntity,
**                           XNodeRef and XDocument.
** 013 CS 20130204           Added conversion support for socket and 
**                           server-socket.
** 014 VMN 20130207          Added unwrapBuffer(), unwrapQuery(), unwrapFieldInterface()
** 015 OM  20130129          Added TempTable handle unwrapper.
** 016 EVL 20130208          Adding wrap methods to invoke SAX object method calls. The new
**                           supported class: XCommon, SaxEntity, SaxAttributes, SaxReader and
**                           SaxWriter.
** 017 CA  20130216          Added unwrappers for SOAP-related handles.
** 018 CS  20130218          Added conversion support for app server.
** 019 VMN 20130219          Added support for attributes
**                           ERROR (KW_ERROR)
**                           NAME (KW_NAME)
**                           PRIVATE-DATA (KW_PRIV_DATA)
**                           DATA-TYPE (KW_DATATYPE)
**                           LABEL (KW_LABEL)
**                           DBNAME (KW_DBNAME)
**                           TABLE (KW_TABLE)
**                           WIDTH-CHARS (KW_WIDTH_C)
**                           HEIGHT-CHARS (KW_HEIGHT_C)
**                           WIDTH-PIXELS (KW_WIDTH_P)
**                           HEIGHT-PIXELS (KW_HEIGHT_P)
** 020 CA  20130221          Reordered unwrap APIs. getType was renamed to getResourceType. Added
**                           unwrap APIs for NAMESPACE-URI, UNIQUE-ID, ADM-DATA, CommonField and 
**                           Sizeable.
** 021 OM  20130220          Added dereference static methods for :: punctuation conversion.
** 022 OM  20130304          Refactored isValid and isUnknown of WrappedResource to valid 
**                           and unknown.
** 023 OM  20130320          Added ResultListHandler to the list of unwrapped interfaces.
** 024 CS  20130321          Added ConnectableServer to the list of unwrapped interfaces.
** 025 GES 20130322          Fixed the BDT c'tor and modified the assign(BDT,boolean) to be the
**                           common location for the _POLY morphing. This is safe because all
**                           current usage would have generated a ClassCastException if that
**                           method was ever used with an invalid type. Now, instead of the CCE,
**                           the code will morph the value as is possible. The morphing code
**                           still needs to be implemented.
** 026 CA  20130529          Remove procedure-related dependencies, exposed the wrapped resource
**                           via a getter.
** 027 OM  20130618          Use default constructor for unknown values of date. Cleanup.
** 028 CA  20130813          Added support for the async request handle.
** 029 CA  20130908          When obtaining the resource ID via resourceId(), make sure the 
**                           resource can be "restored" using the ID.
** 030 CA  20130919          Resource cleanup must be done after it was deleted.
** 031 EVK 20130909          Adding wrap methods to invoke DOM XML object method calls. The
**                           new supported class: XmlSchema, Encoding. Make readOnlyError do not
**                           set Error flag
** 032 CA  20130930          Added c'tor to instantiate a handle from a frame.
** 033 EVL 20131014          Changed the getting attribute name approach to be based on annotation
**                           approach.
** 034 CA  20131021          Adds unwrapper for the DYNAMIC attribute.
** 035 CA  20131028          Fixed error message when a method is called on an invalid handle.
** 036 HC  20131031          Changed resource ID data type to Long, IDs are allocated instead of 
**                           derived from progress resources' hash codes, cleanup of allocated 
**                           resources, see issue #2183.
** 037 CA  20131013          Added INSTANTIATING-PROCEDURE support. Fixed assignment of frame 
**                           resources (as GenericFrame is no longer the legacy resource).
** 038 EVK 20131130          Added util method isHandleTypeOf. Fixed javadoc issues
** 039 CA  20140120          Fixed an error for APPLY statement using an unknown handle.
** 040 VMN 20140123          Cosmetic change to fit to 4GL message.
** 041 CA  20140218          Error messages use the program name as it was used at the RUN call.
** 042 CA  20140326          Added knownResource.
** 043 VIG 20140314          Added unwrapIterableResource() method to support IterableResource.          
** 044 CA  20140526          unwrap APIs were switched to instance method calls.
** 045 CA  20140513          Added a weight for the context-local var, to ensure predetermined 
**                           order during context reset.
** 046 MAG 20140602          Fixed the BDT c'tor and modified the assign(BDT,boolean) on _POLY
**                           morphing. Improved compareTo by comparing this handle with other BDT
**                           types. Added a method to get the resource id if any as Long value.
** 047 VMN 20140702          Fixed InstantiationException in InvalidAttributeAccess.invoke().
** 048 CA  20140703          Added unwrapFontTable.
** 049 MAG 20140812          Added unwrapColorTable, unwrapNumEntries.
** 050 VIG 20140709          Set code2 var in InvalidAttributeAccess subclass.
** 051 SVL 20141102          Added some browse-related unwrap* functions.
** 052 GES 20141028          Added LegacyLogManager support.
** 053 GES 20141124          Added unwrapping for EditorInterface, InnerLines, QueryAssociable,
**                           NumItems and ScrollbarHorizontalElement.
** 054 CA  20150122          APPLY ... TO <handle> is converted to LogicalTerminal APIs (code
**                           related to it was removed).
** 055 EVL 20150128          Adding unwrapping for ImageSupport, ImageButtonSupport interfaces.
** 056 IAS 20150205          Adding unwrapping for ConrolEntity interface.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;

import com.goldencode.p2j.classloader.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.xml.*;

/**
 * A class that represents a Progress 4GL compatible handle data type. This
 * is a container for an object of type {@link WrappedResource}.  Such
 * objects can be passed around as a regular {@link BaseDataType} wrapper
 * but the backing object can be indirectly accessed via the {@link #get} 
 * method.
 * <p>
 * The following is the mapping of Progress language features to the
 * corresponding feature in this class:
 * <p>
 * <pre>
 * string        function (default format)      {@link #toString()}
 *                        (user-defined format) {@link #toString(String)} 
 *                        (export format)       {@link #toStringExport()}
 * valid-handle  function                       {@link #isValid}
 * widget-handle function                       {@link #fromString}
 * </pre>
 * <p>
 * There are 2 deviations between this class and its behavior in the Progress
 * 4GL:
 * <p>
 * <ol>
 *    <li> The default format string is larger in Java than in Progress since
 *         the Java implementation uses all possible 32-bit values as its
 *         handle identifiers. This could be made compatible by limiting the
 *         possible values used to those that would fit into the Progress
 *         equivalent. This would probably require that every handle be
 *         registered or assigned a unique value at construction. 
 *    <li> In Progress, it is possible that one could create a loop to 
 *         instantiate all possible handles using the WIDGET-HANDLE function.
 *         This means that one might not have used the STRING function to get
 *         the text representation since that loop could create strings for
 *         all integers in the possible range of handle identifiers. This
 *         seems like a bad practice, but if it is ever used in reality, it
 *         would invalidate the implementation of {@link #fromString} which
 *         depends upon a map maintained during {@link #toString}.  The
 *         solution is the same as for the oversized text representation
 *         problem above (a registry created at construction of every 
 *         handle).
 * </ol>
 */
public class handle
extends BaseDataType
{
   /** Stores context-local state variables. */
   private static final ContextLocal<WorkArea> work =
   new ContextLocal<WorkArea>()
   {
      @Override
      public WeightFactor getWeight()
      {
         return WeightFactor.LAST;
      };
      
      protected synchronized WorkArea initialValue()
      {
         return new WorkArea();
      }
   };
   
   /** 
    * Dynamic proxy returned by unwrap APIs when the handle is invalid or not
    * of the expected type.  Gets populated on class load.
    */
   private static HandleCommon invalidAttrAccessProxy = null;

   /** Stores the referenced data of this handle. */
   private WrappedResource value = null; 
   
   static
   {
      ClassLoader cl = MultiClassLoader.class.getClassLoader();
      Class<?>[] ifaces = new Class<?>[] { HandleCommon.class };
      InvalidAttributeAccess handler = new InvalidAttributeAccess();
      invalidAttrAccessProxy = (HandleCommon) Proxy.newProxyInstance(cl, ifaces, handler);
   }
   
   /**
    * Default constructor, creates an instance that represents the unknown
    * value.
    */
   public handle()
   {
   }

   /**
    * This is a special c'tor which should be used only when converting the
    * value returned by a function or method with polymorphic return type into the
    * expected type (i.e. DYNAMIC-FUNCTION()).  In such cases, the 4GL does some
    * automatic type conversion (see {@link #assign(BaseDataType,boolean)}).
    * 
    * @param    value
    *           The value to be used for this instance.
    */
   public handle(BaseDataType value)
   {
      if (value == null || value.isUnknown())
      {
         setUnknown();
      }
      else
      {
         assign(value, false);
      }
   }

   /**
    * Constructs an instance that has the contained object reference and
    * <code>unknown value</code> that exactly matches that of the passed-in 
    * instance.
    *
    * @param    value
    *           The instance to pattern from.
    */
   public handle(handle value)
   {
      assign(value, false);
   }
   
   /**
    * Constructs an instance that has the contained object reference.
    *
    * @param    value
    *           The instance to pattern from.
    */
   public handle(WrappedResource value)
   {
      assign(value, false);
   }
   
   /**
    * Constructs an instance that has the contained object reference.
    *
    * @param    value
    *           The frame instance.
    */
   public handle(CommonFrame value)
   {
      assign(value.asWidgetHandle(), false);
   }

   /**
    * Check the type of resource in the handle. Return {@code true} if and only if handle
    * is valid and contains resource of specific type.
    *
    * @param    handleToCheck
    *           The handle instance to check.
    * @param    clazz
    *           Specific class type which will be compare.
    *
    * @return   {@code true} if handleToCheck contains resource an instance of clazz,
    *           {@code false} otherwise.
    */
   public static boolean isHandleTypeOf(handle handleToCheck, Class<?> clazz)
   {
      if (handleToCheck == null || !handleToCheck._isValid() || handleToCheck.isUnknown())
      {
         return false;
      }
      WrappedResource resource = handleToCheck.getResource();

      return clazz.isInstance(resource);
   }

   /**
    * Emitted and to be used only in cases the attribute is read-only. It
    * displays appropriate error message.
    * 
    * @param    h
    *           The handle used to access the attribute.
    * @param    attribute
    *           The read-only attribute.
    */
   public static void readOnlyError(handle h, String attribute)
   {
      if (attribute == null)
      {
         throw new IllegalArgumentException("The attribute is null !");
      }

      final String msg = "**%s is not a setable attribute for %s widget";
      String err = String.format(msg,
                                 attribute.toUpperCase(),
                                 h.unwrap().getResourceType().toStringMessage());
      ErrorManager.recordOrShowError(4052, err, false, false, false);
   }

   /**
    * Convert the string representation of a handle into the original handle
    * instance. This facility can only work for handles that have been
    * converted at least once into a string via {@link #toString}. Such
    * handles are stored in a map for conversion purposes.
    *
    * @param    txt
    *           A valid text representation of a handle.
    */
   public static handle fromString(character txt)
   {
      return fromString(txt.toStringMessage());
   }
   
   /**
    * Convert the string representation of a handle into the original handle
    * instance. This facility can only work for handles that have been
    * converted at least once into a string via {@link #toString}. Such
    * handles are stored in a map for conversion purposes.
    *
    * @param    txt
    *           A valid text representation of a handle.
    */
   public static handle fromString(String txt)
   {
      Long id = ResourceIdHelper.idFromString(txt);
      if (ResourceIdHelper.isUnknown(id))
      {
         return new handle();
      } 
      else if (ResourceIdHelper.isZero(id))
      {
         return new handle(ZERO_RESOURCE);
      }
      
      return fromResourceId(id);
   }
   
   /**
    * Converts the number representation of a handle into the original handle
    * instance. 
    * <p>
    * For handles that have been converted at least once into a string via 
    * {@link #toString} the method returns a valid handle instance. For handles
    * that have not been converted at least once onto a string via {@link #toString}
    * the method returns a zero handle, i.e. handle with "0" string representation. 
    * 
    * @param   resourceId
    *          Resource identifier. 
    */
   public static handle fromResourceId(Long resourceId)
   {
      WorkArea wa = work.get();
      WrappedResource resource = wa.idResources.get(resourceId);
      
      if (resource != null)
      {
         return new handle(resource);
      }
      
      return new handle(ZERO_RESOURCE);
   }
   
   /**
    * This API should be called only in case of a attribute/method access using a resource which
    * does not support it.  It shows appropriate error messages, depending on handle state and
    * attribute/method.
    * <p>
    * This must not be called in cases of a not-supported attribute invoked as a lvalue (i.e. a
    * setter for a not-supported attribute is invoked). In this case, call {@link #readOnlyError}.
    * 
    * @param    attr
    *           The attribute/method name which is not supported by the handle.
    * @param    h
    *           The handle used to access the attribute/method.
    */
   public static void invalidAttribute(String attr, handle h)
   {
      if (!h._isValid())
      {
         final String errMsg1 = 
            "Invalid handle.  Not initialized or points to a deleted " +
            "object";
         final String errMsg2 = 
           "Cannot access the %s attribute because the widget does not " +
           "exist";
         int[] errIds = { 3135, 3140 };
         String[] msgs = 
         {
            errMsg1,
            String.format(errMsg2, attr)
         };
         
         ErrorManager.recordOrThrowError(errIds, msgs, false);
      }
      else
      {
         // valid handle, wrong API, show error
         final String errMsg = 
            "**%s is not a queryable attribute for %s widget";

         String type = h.unwrapType().getResourceType().toStringMessage();
         String msg = String.format(errMsg, attr, type);
         ErrorManager.displayError(4052, msg, false);
      }
   }
   
   /**
    * Remove the given resource from the context-local map of all resource.
    * <p>
    * This method should be called by all wrapped resources when deleted,
    * that is all resources implementing {@link WrappedResource} interface.  
    * 
    * @param    resource
    *           A resource instance to be removed.
    */
   public static void removeResource(WrappedResource resource)
   {
      if (resource == null)
      {
         return;
      }
      
      WorkArea wa = work.get();
      Long id = wa.resourceIds.get(resource);
      if (id != null)
      {
         wa.idResources.remove(id);
         wa.resourceIds.remove(resource);
      }
   }

   /**
    * Get the ID of the given resource, as used by the {@link #toString} and
    * {@link #fromString} APIs.
    * <p>
    * The method allocates additional resources together with the provided wrapped resource, 
    * thus it is necessary that {@link #removeResource(WrappedResource)} is called when 
    * the returned resource id is not needed anymore (i.e. the related resource is deleted),
    * see {@link #removeResource(WrappedResource)} for more details.
    * 
    * @param    resource
    *           A resource instance.
    *
    * @return   The ID of this resource.
    */
   public static Long resourceId(WrappedResource resource)
   {
      if (resource == ZERO_RESOURCE)
      {
         return 0l;
      }
      
      WorkArea wa = work.get();
      
      // resource id already allocated?
      Long existingId = wa.resourceIds.get(resource);
      if (existingId != null)
      {
         // yes
         return existingId;
      }
      
      // allocate new resource id - collision with existing ids must be prevented
      Long id = null;
      do 
      {
         id = ResourceIdHelper.getNext();
      } 
      while (wa.idResources.containsKey(id));
   
      wa.idResources.put(id, resource);
      wa.resourceIds.put(resource, id);
   
      return id;
   }

   /**
    * Check if the given resource is known (has an associated ID).
    * 
    * @param    resource
    *           A resource, valid or not.
    * 
    * @return   {@code true} if the resource has an associated ID.
    */
   static boolean knownResource(WrappedResource resource)
   {
      WorkArea wa = work.get();
      
      return wa.resourceIds.get(resource) != null;
   }

   /**
    * Unwrap the given handle so that it returns an instance compatible with
    * the given interface, <code>ifc</code>. If the handle is invalid or is
    * not compatible with the given interface, a dynamic proxy is returned
    * (see {@link #invalidAttrAccessProxy}.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @param    h
    *           The handle to be unwrapped.
    * @param    ifc
    *           The expected interface.
    *            
    * @return   See above.
    */
   @SuppressWarnings("unchecked")
   private static <T> T unwrapImpl(handle h, Class<T> ifc)
   {
      WrappedResource value = h.value;
      if (h._isValid() && ifc.isAssignableFrom(value.getClass()))
      {
         return (T) value;
      }
   
      work.get().invalidCalls.push(h);
      return (T) invalidAttrAccessProxy;
   }

   /**
    * Get the legacy attribute or method name mapped by this method.
    * 
    * @param    method
    *           The method for which the legacy name is needed.
    * @param    args
    *           The method's arguments.
    * 
    * @return   The legacy name from the method annotation or the converted API name, if not
    *           found in the respective annotation.
    */
   private static String getAttrOrMethodName(Method method, Object[] args)
   {
      String name = method.getName();
   
      if ("readOnlyError".equals(name) &&
          args.length == 1             &&
          args[0] != null              &&
          (args[0] instanceof String)  &&
          method.getDeclaringClass().equals(CommonHandle.class))
      {
         return ((String) args[0]).toUpperCase();
      }
      
      String api = null;
      LegacyAttribute legacyAttr = method.getAnnotation(LegacyAttribute.class);
      if (legacyAttr != null && !legacyAttr.ignore())
      {
         api = legacyAttr.name();
      }
            
      LegacyMethod legacyMthd = method.getAnnotation(LegacyMethod.class);
      if (legacyMthd != null)
      {
         api = legacyMthd.name();
      }
   
      return api == null ? name : api;
   }

   /**
    * Unwrap this handle to a {@link CommonHandleChain} instance. If the
    * handle is not a procedure, widget, window or frame handle or is invalid, 
    * an {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    *           
    * @return   See above.
    */
   public CommonHandleChain unwrap()
   {
      return unwrapImpl(this, CommonHandleChain.class);
   }

   /**
    * Unwrap this handle to a {@link ADMData} instance. If the handle is 
    * not a ADMData handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @return   See above.
    */
   public ADMData unwrapADMData()
   {
      return unwrapImpl(this, ADMData.class);
   }

   /**
    * Unwrap this handle to a {@link AsyncRequest} instance. If the handle is 
    * not a <code>AsynchronousRequest</code> handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @return   See above.
    */
   public AsyncRequest unwrapAsync()
   {
      return unwrapImpl(this, AsyncRequest.class);
   }

   /**
    * Unwrap this handle to a {@link AsyncRequestCountAttribute} instance. If the handle is 
    * not a AsyncRequestCountAttribute handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @return   See above.
    */
   public AsyncRequestCountAttribute unwrapAsyncRequestCount()
   {
      return unwrapImpl(this, AsyncRequestCountAttribute.class);
   }

   /**
    * Unwrap this handle to a {@link Buffer} instance. If the
    * handle is not the Buffer handle or is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle.
    *
    * @return   See above.
    */
   public Buffer unwrapBuffer()
   {
      return unwrapImpl(this, Buffer.class);
   }

   /**
    * Unwrap this handle to a {@link FieldInterface} instance. If the
    * handle is not the FieldInterface handle or is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle.
    *
    * @return   See above.
    */
   public FieldInterface unwrapBufferField()
   {
      return unwrapImpl(this, FieldInterface.class);
   }

   /**
    * Unwrap this handle to a {@link ColorTableResource} instance. If the handle is
    * not a ColorTableResource handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @return   See above.
    */
   public ColorTableResource unwrapColorTable()
   {
      return unwrapImpl(this, ColorTableResource.class);
   }

   /**
    * Unwrap this handle to an {@link com.goldencode.p2j.ui.EditorInterface} instance. If the
    * handle is not an <code>EditorInterface</code> handle or is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is
    * needed because for invalid handles, a stack of handles is maintained, to know the state
    * of the last used handle.
    *
    * @return   See above.
    */
   public EditorInterface unwrapEditor()
   {
      return unwrapImpl(this, EditorInterface.class);
   }

   /**
    * Unwrap this handle to a {@link com.goldencode.p2j.ui.BrowseInterface} instance. If the
    * handle is not the <code>BrowseInterface</code> handle or is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle.
    *
    * @return   See above.
    */
   public BrowseInterface unwrapBrowse()
   {
      return unwrapImpl(this, BrowseInterface.class);
   }

   /**
    * Unwrap this handle to an {@link com.goldencode.p2j.ui.FrameInterface} instance. If the
    * handle is not an <code>FrameInterface</code> handle or is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is
    * needed because for invalid handles, a stack of handles is maintained, to know the state
    * of the last used handle.
    *
    * @return   See above.
    */
   public FrameInterface unwrapFrame()
   {
      return unwrapImpl(this, FrameInterface.class);
   }

   /**
    * Unwrap this handle to a {@link AutoZapElement} instance. If the
    * handle is not the AutoZapElement handle or is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle.
    *
    * @return   See above.
    */
   public AutoZapElement unwrapAutoZapElement()
   {
      return unwrapImpl(this, AutoZapElement.class);
   }

   /**
    * Unwrap this handle to a {@link ScrollbarHorizontalElement} instance. If the
    * handle is not the <code>ScrollbarHorizontalElement</code> handle or is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle.
    *
    * @return   See above.
    */
   public ScrollbarHorizontalElement unwrapScrollbarHorizontalElement()
   {
      return unwrapImpl(this, ScrollbarHorizontalElement.class);
   }

   /**
    * Unwrap this handle to a {@link ScrollbarVerticalElement} instance. If the
    * handle is not the ScrollbarVerticalElement handle or is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle.
    *
    * @return   See above.
    */
   public ScrollbarVerticalElement unwrapScrollbarVerticalElement()
   {
      return unwrapImpl(this, ScrollbarVerticalElement.class);
   }

   /**
    * Unwrap this handle to a {@link BrowseColumnInterface} instance. If the
    * handle is not the BrowseColumnInterface handle or is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle.
    *
    * @return   See above.
    */
   public BrowseColumnInterface unwrapBrowseColumn()
   {
      return unwrapImpl(this, BrowseColumnInterface.class);
   }

   /**
    * Unwrap this handle to a {@link BrowseElement} instance. If the
    * handle is not the BrowseElement handle or is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle.
    *
    * @return   See above.
    */
   public BrowseElement unwrapBrowseElement()
   {
      return unwrapImpl(this, BrowseElement.class);
   }

   /**
    * Unwrap this handle to a {@link InnerLines} instance. If the
    * handle is not the <code>InnerLines</code> handle or is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle.
    *
    * @return   See above.
    */
   public InnerLines unwrapInnerLines()
   {
      return unwrapImpl(this, InnerLines.class);
   }

   /**
    * Unwrap this handle to a {@link MaxCharsElement} instance. If the
    * handle is not the MaxCharsElement handle or is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle.
    *
    * @return   See above.
    */
   public MaxCharsElement unwrapMaxCharsElement()
   {
      return unwrapImpl(this, MaxCharsElement.class);
   }

   /**
    * Unwrap this handle to a {@link SortedElement} instance. If the
    * handle is not the SortedElement handle or is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle.
    *
    * @return   See above.
    */
   public SortedElement unwrapSortedElement()
   {
      return unwrapImpl(this, SortedElement.class);
   }

   /**
    * Unwrap this handle to a {@link CommonProcedure} instance. If the
    * handle is not a procedure or the SESSION system handle or is invalid, 
    * an {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public CommonProcedure unwrapCommonProcedure()
   {
      return unwrapImpl(this, CommonProcedure.class);
   }

   /**
    * Unwrap this handle to a {@link CommonField} instance. If the
    * handle is not CommonField resource or is invalid, 
    * an {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public CommonField unwrapCommonField()
   {
      return unwrapImpl(this, CommonField.class);
   }

   /**
    * Unwrap this handle to a {@link Connectable} instance. If the
    * handle is not a Connectable handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public Connectable unwrapConnectable()
   {
      return unwrapImpl(this, Connectable.class);
   }
   
   /**
    * Unwrap this handle to a {@link ConnectableServer} instance. If the
    * handle is not a ConnectableServer handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public ConnectableServer unwrapConnectableServer()
   {
      return unwrapImpl(this, ConnectableServer.class);
   }

   /**
    * Unwrap this handle to a {@link DatabaseInfo} instance. If the
    * handle is not a procedure, widget, window or frame handle or is invalid,
    * an {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle.
    *
    * @return   See above.
    */
   public DatabaseInfo unwrapDatabaseInfo()
   {
      return unwrapImpl(this, DatabaseInfo.class);
   }

   /**
    * Unwrap this handle to a {@link Dereferenceable} instance. If the
    * handle is not the Dereferenceable handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation to dereference().  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public Dereferenceable unwrapDereferenceable()
   {
      return unwrapImpl(this, Dereferenceable.class);
   }

   /**
    * Unwrap this handle to a {@link DynamicResource} instance. If the
    * handle is not the DynamicResource handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation to dereference().  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public DynamicResource unwrapDynamicResource()
   {
      return unwrapImpl(this, DynamicResource.class);
   }

   /**
    * Unwrap this handle to a {@link XmlSchema} instance. If the
    * handle is invalid, an {@link #invalidAttrAccessProxy dynamic proxy}
    * is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle.
    *
    * @return   See above.
    */
   public Encoding unwrapEncoding()
   {
      return unwrapImpl(this, Encoding.class);
   }

   /**
    * Unwrap this handle to a {@link Errorable} instance. If the
    * handle is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle.
    *
    * @return   See above.
    */
   public Errorable unwrapError()
   {
      return unwrapImpl(this, Errorable.class);
   }

   /**
    * Unwrap this handle to a {@link CommonErrorStatus} instance. If the
    * handle is not the ERROR-STATUS handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public CommonErrorStatus unwrapErrorStatus()
   {
      return unwrapImpl(this, CommonErrorStatus.class);
   }
   
   /**
    * Unwrap this handle to a {@link CommonFile} instance. If the
    * handle is not the FILE-INFO or a procedure handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public CommonFile unwrapFile()
   {
      return unwrapImpl(this, CommonFile.class);
   }

   /**
    * Unwrap this handle to a {@link CommonFileInfo} instance. If the
    * handle is not the FILE-INFO handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public CommonFileInfo unwrapFileInfo()
   {
      return unwrapImpl(this, CommonFileInfo.class);
   }
   
   /**
    * Unwrap this handle to a {@link FirstLastProcedureAttribute} instance. If the handle is
    * not a FirstLastProcedureAttribute handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @return   See above.
    */
   public FirstLastProcedureAttribute unwrapFirstLastProcedure()
   {
      return unwrapImpl(this, FirstLastProcedureAttribute.class);
   }
   
   /**
    * Unwrap this handle to a {@link FontTableResource} instance. If the handle is
    * not a FontTableResource handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @return   See above.
    */
   public FontTableResource unwrapFontTable()
   {
      return unwrapImpl(this, FontTableResource.class);
   }
   
   /**
    * Unwrap this handle to a {@link InstantiatingProcedure} instance. If the handle is
    * not a FirstLastProcedureAttribute handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @return   See above.
    */
   public InstantiatingProcedure unwrapInstantiatingProcedure()
   {
      return unwrapImpl(this, InstantiatingProcedure.class);
   }
   
   /**
    * Unwrap this handle to a {@link NumEntries} instance. If the handle is
    * not a NumEntries handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @return   See above.
    */
   public NumEntries unwrapNumEntries()
   {
      return unwrapImpl(this, NumEntries.class);
   }

   /**
    * Unwrap this handle to a {@link NumItems} instance. If the handle is not a NumItems handle
    * or is invalid, an {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @return   See above.
    */
   public NumItems unwrapNumItems()
   {
      return unwrapImpl(this, NumItems.class);
   }

   /**
    * Unwrap this handle to a {@link ImageButtonSupport} instance. If the handle is not
    * a ImageButtonSupport handle or is invalid, an {@link #invalidAttrAccessProxy dynamic proxy}
    * is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @return   See above.
    */
   public ImageButtonSupport unwrapImageButtonSupport()
   {
      return unwrapImpl(this, ImageButtonSupport.class);
   }

   /**
    * Unwrap this handle to a {@link ImageSupport} instance. If the handle is not a ImageSupport
    * handle or is invalid, an {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @return   See above.
    */
   public ImageSupport unwrapImageSupport()
   {
      return unwrapImpl(this, ImageSupport.class);
   }

   /**
    * Unwrap this handle to a {@link IterableResource} instance. 
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @return   See above.
    */
   public IterableResource unwrapIterableResource()
   {
      return unwrapImpl(this, IterableResource.class);
   }
   
   /**
    * Unwrap this handle to a {@link CommonLastEvent} instance. If the
    * handle is not the KEY-READER handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public CommonLastEvent unwrapLastEvent()
   {
      return unwrapImpl(this, CommonLastEvent.class);
   }

   /**
    * Unwrap this handle to a {@link LegacyLogManager} instance. If the
    * handle is not the LOG-MANAGER or DSLOG-MANAGER handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public LegacyLogManager unwrapLogManager()
   {
      return unwrapImpl(this, LegacyLogManager.class);
   }

   /**
    * Unwrap this handle to a {@link LocalName} instance. If the handle is 
    * not a LocalName handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @return   See above.
    */
   public LocalName unwrapLocalName()
   {
      return unwrapImpl(this, LocalName.class);
   }

   /**
    * Unwrap this handle to a {@link NamespaceURI} instance. If the handle is 
    * not a NamespaceURI handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @return   See above.
    */
   public NamespaceURI unwrapNamespaceURI()
   {
      return unwrapImpl(this, NamespaceURI.class);
   }

   /**
    * Unwrap this handle to a {@link PersistentProcedure} instance. If 
    * the handle is not a procedure handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public PersistentProcedure unwrapProcedure()
   {
      return unwrapImpl(this, PersistentProcedure.class);
   }

   /**
    * Unwrap this handle to a {@link P2JQuery} instance. If the
    * handle is not a query handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public P2JQuery unwrapQuery()
   {
      return unwrapImpl(this, P2JQuery.class);
   }
   
   /**
    * Unwrap this handle to a {@link QueryAssociable} instance. If the
    * handle is not a query handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public QueryAssociable unwrapQueryAssociable()
   {
      return unwrapImpl(this, QueryAssociable.class);
   }
   
   /**
    * Unwrap this handle to a {@link Remotable} instance. If the
    * handle is not a <code>Remotable</code> handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public Remotable unwrapRemotable()
   {
      return unwrapImpl(this, Remotable.class);
   }
   
   /**
    * Unwrap this handle to a {@link ResultListHandler} instance. If the
    * handle is not a <code>ResultListHandler</code> handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public ResultListHandler unwrapResultListHandler()
   {
      return unwrapImpl(this, ResultListHandler.class);
   }

   /**
    * Unwrap this handle to a {@link SaxAttributes} instance. If the handle is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is
    * needed because for invalid handles, a stack of handles is maintained, to know the state of
    * the last used handle. 
    * 
    * @return   See above.
    */
   public SaxAttributes unwrapSaxAttributes()
   {
      return unwrapImpl(this, SaxAttributes.class);
   }

   /**
    * Unwrap this handle to a {@link SaxEntity} instance. If the handle is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is
    * needed because for invalid handles, a stack of handles is maintained, to know the state of
    * the last used handle. 
    * 
    * @return   See above.
    */
   public SaxEntity unwrapSaxEntity()
   {
      return unwrapImpl(this, SaxEntity.class);
   }

   /**
    * Unwrap this handle to a {@link SaxReader} instance. If the handle is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is
    * needed because for invalid handles, a stack of handles is maintained, to know the state of
    * the last used handle. 
    * 
    * @return   See above.
    */
   public SaxReader unwrapSaxReader()
   {
      return unwrapImpl(this, SaxReader.class);
   }

   /**
    * Unwrap this handle to a {@link SaxWriter} instance. If the handle is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is
    * needed because for invalid handles, a stack of handles is maintained, to know the state of
    * the last used handle. 
    * 
    * @return   See above.
    */
   public SaxWriter unwrapSaxWriter()
   {
      return unwrapImpl(this, SaxWriter.class);
   }

   /**
    * Unwrap this handle to a {@link Sensitive} instance. If the
    * handle is not a Sensitive handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public Sensitive unwrapSensitive()
   {
      return unwrapImpl(this, Sensitive.class);
   }

   /**
    * Unwrap this handle to a {@link Server} instance. If the handle is not a Server handle 
    * or is invalid, an {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @return   See above.
    */
   public Server unwrapServer()
   {
      return unwrapImpl(this, Server.class);
   }

   /**
    * Unwrap this handle to a {@link CommonProcedure} instance. If the
    * handle is not the SESSION system handle or is invalid, 
    * an {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public CommonSession unwrapSession()
   {
      return unwrapImpl(this, CommonSession.class);
   }

   /**
    * Unwrap this handle to a {@link Sizeable} instance. If the
    * handle is not a procedure, widget, window or frame handle or is invalid,
    * an {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle.
    *
    * @return   See above.
    */
   public Sizeable unwrapSizeable()
   {
      return unwrapImpl(this, Sizeable.class);
   }

   /**
    * Unwrap this handle to a {@link SOAPEntity} instance. If the handle's resource is 
    * not implementing the {@link SOAPEntity} interface or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @return   See above.
    */
   public SOAPEntity unwrapSOAPEntity()
   {
      return unwrapImpl(this, SOAPEntity.class);
   }

   /**
    * Unwrap this handle to a {@link SOAPFault} instance. If the handle's resource is 
    * not implementing the {@link SOAPFault} interface or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @return   See above.
    */
   public SOAPFault unwrapSOAPFault()
   {
      return unwrapImpl(this, SOAPFault.class);
   }

   /**
    * Unwrap this handle to a {@link SOAPHeader} instance. If the handle's resource is 
    * not implementing the {@link SOAPHeader} interface or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @return   See above.
    */
   public SOAPHeader unwrapSOAPHeader()
   {
      return unwrapImpl(this, SOAPHeader.class);
   }

   /**
    * Unwrap this handle to a {@link SOAPHeaderEntry} instance. If the handle's resource is 
    * not implementing the {@link SOAPHeaderEntry} interface or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @return   See above.
    */
   public SOAPHeaderEntry unwrapSOAPHeaderEntry()
   {
      return unwrapImpl(this, SOAPHeaderEntry.class);
   }

   /**
    * Unwrap this handle to a {@link Socket} instance. If the
    * handle is not the SOCKET handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public Socket unwrapSocket()
   {
      return unwrapImpl(this, Socket.class);
   }

   /**
    * Unwrap this handle to a {@link SocketListener} instance. 
    * If the handle is not the SERVER-SOCKET handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public SocketListener unwrapSocketListener()
   {
      return unwrapImpl(this, SocketListener.class);
   }

   /**
    * Unwrap this handle to a {@link SslServerAttribute} instance. If the handle is not a 
    * SslServerAttribute handle or is invalid, an {@link #invalidAttrAccessProxy dynamic proxy}
    * is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @return   See above.
    */
   public SslServerAttribute unwrapSslServerAttribute()
   {
      return unwrapImpl(this, SslServerAttribute.class);
   }

   /**
    * Unwrap this handle to a {@link SubTypeAttribute} instance. If the handle is not a 
    * SubTypeAttribute handle or is invalid, an {@link #invalidAttrAccessProxy dynamic proxy} 
    * is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @return   See above.
    */
   public SubTypeAttribute unwrapSubType()
   {
      return unwrapImpl(this, SubTypeAttribute.class);
   }

   /**
    * Unwrap this handle to a {@link TempTable} instance. If the
    * handle is not a temporary handle or is invalid, 
    * an {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public TempTable unwrapTempTable()
   {
      return unwrapImpl(this, TempTable.class);
   }

   /**
    * Unwrap this handle to a {@link CommonHandle} instance. If the
    * handle is invalid, an {@link #invalidAttrAccessProxy dynamic proxy} 
    * is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public CommonHandle unwrapType()
   {
      return unwrapImpl(this, CommonHandle.class);
   }

   /**
    * Unwrap this handle to a {@link UniqueID} instance. If the handle is 
    * not a UniqueID handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is needed
    * because for invalid handles, a stack of handles is maintained, to know the state of the last
    * used handle. 
    * 
    * @return   See above.
    */
   public UniqueID unwrapUniqueID()
   {
      return unwrapImpl(this, UniqueID.class);
   }

   /**
    * Unwrap this handle to a {@link XCommon} instance. If the handle is invalid, an
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method invocation.  This is
    * needed because for invalid handles, a stack of handles is maintained, to know the state of
    * the last used handle. 
    * 
    * @return   See above.
    */
   public XCommon unwrapXCommon()
   {
      return unwrapImpl(this, XCommon.class);
   }

   /**
    * Unwrap this handle to a {@link XEntity} instance. If the
    * handle is invalid, an {@link #invalidAttrAccessProxy dynamic proxy} 
    * is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public XEntity unwrapXEntity()
   {
      return unwrapImpl(this, XEntity.class);
   }

   /**
    * Unwrap this handle to a {@link XDocument} instance. If the
    * handle is invalid, an {@link #invalidAttrAccessProxy dynamic proxy} 
    * is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public XDocument unwrapXDocument()
   {
      return unwrapImpl(this, XDocument.class);
   }

   /**
    * Unwrap this handle to a {@link XmlSchema} instance. If the
    * handle is invalid, an {@link #invalidAttrAccessProxy dynamic proxy}
    * is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle.
    *
    * @return   See above.
    */
   public XmlSchema unwrapXmlSchema()
   {
      return unwrapImpl(this, XmlSchema.class);
   }

   /**
    * Unwrap this handle to a {@link XNodeRef} instance. If the
    * handle is invalid, an {@link #invalidAttrAccessProxy dynamic proxy} 
    * is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public XNodeRef unwrapXNodeRef()
   {
      return unwrapImpl(this, XNodeRef.class);
   }

   /**
    * Unwrap this handle to a {@link CommonWidget} instance. If the
    * handle is not a widget handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public CommonWidget unwrapWidget()
   {
      return unwrapImpl(this, CommonWidget.class);
   }

   /**
    * Unwrap this handle to a {@link CommonWidget} instance. If the
    * handle is not a widget handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public CommonListWidget unwrapCommonListWidget()
   {
      return unwrapImpl(this, CommonListWidget.class);
   }

   /**
    * Unwrap this handle to a {@link RadioSetInterface} instance. If the
    * handle is not a widget handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public RadioSetInterface unwrapRadioSetInterface()
   {
      return unwrapImpl(this, RadioSetInterface.class);
   }

  /**
    * Unwrap this handle to a {@link ControlEntity} instance. If the
    * handle is not a widget handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
  public ControlEntity unwrapControlEntity()
   {
      return unwrapImpl(this, ControlEntity.class);
   }

   /**
    * Unwrap this handle to a {@link CommonWindow} instance. If the
    * handle is a window handle or is invalid, an 
    * {@link #invalidAttrAccessProxy dynamic proxy} is returned.
    * <p>
    * Each unwrap call must be followed by an actual attribute/method 
    * invocation.  This is needed because for invalid handles, a stack of
    * handles is maintained, to know the state of the last used handle. 
    * 
    * @return   See above.
    */
   public CommonWindow unwrapWindow()
   {
      return unwrapImpl(this, CommonWindow.class);
   }
   
   /**
    * Creates a new instance of the same type that represents the 
    * <code>unknown value</code>.
    *
    * @return   An instance that represents the <code>unknown value</code>.
    */
   public BaseDataType instantiateUnknown()   
   {
      return new handle();
   }
   
   /**
    * Does the same as standard <code>clone()</code> method but returns an
    * instance of <code>BaseDataType</code> and doesn't throw the
    * <code>CloneNotSupportedException</code>.
    *
    * @return   A clone of this instance.  
    */
   public BaseDataType duplicate()
   {
      return new handle(this);
   }
   
   /**
    * Hash code implementation which is consistent with {@link
    * BaseDataType#equals}.
    *
    * @return  Hash code value for this object instance.
    */
   public int hashCode()
   {  
      if (isUnknown())
      {
         return ResourceIdHelper.UNKNOWN_STRING.hashCode();
      }
      
      Long resId = resourceId(this.value);
      return resId.hashCode();
   }

   /**
    * Returns the value of this instance as an <code>Object</code>.
    * <p>
    * If this instance represents the <code>unknown value</code>, a
    * <code>null</code> will be returned.
    *
    * @return   The value as an <code>Object</code> or <code>null</code> if
    *           this is <code>unknown</code>.
    */
   public Object get()
   {
      Object result = value;
      
      // redirect to contained object if it is the special wrapper
      if (result != null && result instanceof TransparentWrapper)
      {
         TransparentWrapper wrap = (TransparentWrapper) result;
         result = wrap.get();
      }
      
      return result;
   }   
   
   /**
    * Set the value of this instance as a {@link WrappedResource}.
    *
    * @param    value
    *           The value to be represented in this instance.
    */
   public void set(WrappedResource value)
   {
      this.value = value;
   }
   
   /**
    * Determines if this instance is valid (to be used).
    *
    * @return   <code>true</code> if this instance can be used.
    */
   public logical isValid()
   {
      return new logical(_isValid());
   }   
   
   /**
    * Determines if this instance is valid (to be used).
    *
    * @return   boolean <code>true</code> if this instance can be used.
    */
   public boolean _isValid()
   {
      return (value == null) ? false : value.valid();
   }   
   
   /**
    * Determines if this instance represents the <code>unknown value</code>.
    * 
    * @return   <code>true</code> if this instance is set to the
    *           <code>unknown value</code>.
    */
   public boolean isUnknown()
   {
      return (value == null || value.unknown());
   }
   
   /**
    * Sets the state of this instance's <code>unknown value</code> flag to
    * <code>true</code>.
    * <p>
    * <b>Warning: a separate call is needed to ensure that the data of this
    * instance is set to the correct value.</b>
    */
   public void setUnknown()
   {
      value = null;
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    * <p>
    * If the value is not of type <code>handle</code>, the following automatic type
    * conversion will occur:
    * <p>
    * <pre>
    * A handle constructor accept only a handler BDT instance as parameter. 
    * </pre>
    * <p>
    * This conversion is meant to handle the cases of built-in functions and methods
    * in the 4GL which have polymorphic return types (e.g. DYNAMIC-FUNCTION()).
    *
    * @param    value
    *           The instance from which to copy state.
    * @param    force
    *           Flag to indicate if a pending error should be ignored (see 
    *           {@link ErrorManager#isPendingError} and the assignment
    *           forced regardless.  Note that normal assignment silently
    *           exits with no change if there is a pending error. If this
    *           flag is <code>true</code>, that check is bypassed and the
    *           assignment will always be done.
    */
   public void assign(BaseDataType value, boolean force)
   {
      if (value instanceof handle)
      {
         assign((handle) value, force);
      }
      else
      {
         String err = "Incompatible data types in expression or assignment.";
         ErrorManager.recordOrThrowError(223, err);      
      }
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    *
    * @param    value
    *           The instance from which to copy state.
    * @param    force
    *           Flag to indicate if a pending error should be ignored (see 
    *           {@link ErrorManager#isPendingError} and the assignment
    *           forced regardless.  Note that normal assignment silently
    *           exits with no change if there is a pending error. If this
    *           flag is <code>true</code>, that check is bypassed and the
    *           assignment will always be done.
    */
   public void assign(handle value, boolean force)
   {
      assign(value.value, force);
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    *
    * @param    value
    *           The instance from which to copy state.
    * @param    force
    *           Flag to indicate if a pending error should be ignored (see 
    *           {@link ErrorManager#isPendingError} and the assignment
    *           forced regardless.  Note that normal assignment silently
    *           exits with no change if there is a pending error. If this
    *           flag is <code>true</code>, that check is bypassed and the
    *           assignment will always be done.
    */
   public void assign(WrappedResource value, boolean force)
   {
      if (force || !ErrorManager.isPendingError())
      {
         set(value);
      }
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    * <p>
    * This is a "no operation" in the case where silent error mode is active
    * AND an error has occurred in the current processing since the opening
    * call to {@link ErrorManager#silentErrorEnable}. This protection is
    * needed to ensure that the state of this instance does not change in
    * this case.  This will not affect undo because that only occurs when
    * silent error mode is disabled.
    *
    * @param    value
    *           The instance from which to copy state. This must be of
    *           type <code>handle</code> or <code>WrappedResource</code>.
    */
   public void assign(Object value)
   {
      if (value instanceof WrappedResource)
      {
         assign((WrappedResource) value);
      }
      else if (value instanceof handle)
      {
         assign((handle) value);
      }
      else if (value instanceof CommonFrame)
      {
         value = ((CommonFrame) value).asWidget();

         assign((WrappedResource) value);
      }
      else
      {
         // unexpected failure
         throw new IllegalArgumentException("Invalid value type.");
      }
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    * <p>
    * This is a "no operation" in the case where silent error mode is active
    * AND an error has occurred in the current processing since the opening
    * call to {@link ErrorManager#silentErrorEnable}. This protection is
    * needed to ensure that the state of this instance does not change in
    * this case.  This will not affect undo because that only occurs when
    * silent error mode is disabled.
    *
    * @param    value
    *           The instance from which to copy state.
    */
   public void assign(handle value)
   {
      assign(value.value, false);
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    *
    * @param    value
    *           The instance from which to copy state.
    */
   public void assign(Undoable value)
   {
      assign((handle) value, false);
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * given value.  Note that <code>null</code> may be passed in which case
    * this instance will be set to the <code>unknown value</code>.
    * <p>
    * This is a "no operation" in the case where silent error mode is active
    * AND an error has occurred in the current processing since the opening
    * call to {@link ErrorManager#silentErrorEnable}. This protection is
    * needed to ensure that the state of this instance does not change in
    * this case.  This will not affect undo because that only occurs when
    * silent error mode is disabled.
    *
    * @param    value
    *           The instance from which to copy state.
    */
   public void assign(WrappedResource value)
   {
      assign(value, false);
   }
   
   /**
    * Compares this instance with the specified instance and returns a -1
    * if this instance is less than the specified, 0 if the two instances
    * are equal and 1 if this instance is greater than the specified
    * instance.  This is the implementation of the <code>Comparable</code>
    * interface.
    * <p>
    * The algorithm will fail to give meaningful results in the case where
    * one tries to sort against other objects that do not represent compatible
    * values.
    *
    * @param    obj
    *           The instance to compare against.
    *
    * @return   0 or 1 depending on if the passed instance is or is not
    *           <code>unknown</code>.
    */
   public int compareTo(Object obj)   
   {  
      Long bdtValue = null;
      Long resourceId = resourceId(getResource());
            
      // handle is also a BDT type
      if (obj instanceof BaseDataType)
      {         
         BaseDataType bdt = (BaseDataType) obj;

         // two unknown values are equals
         if (isUnknown() || bdt.isUnknown())
         {
            // an unknown handle is greater than a known handle
            return Boolean.compare(isUnknown(), bdt.isUnknown());
         }

         if (bdt instanceof handle)
         {
            // handle
            bdtValue = resourceId(((handle) bdt).getResource());
         }
         else
         {
            // this path is only used when type morphing is needed         
            // ResourceIdHelper.idFromString could throw an ErrorConditionException
            try
            {
               if (bdt instanceof date && !(bdt instanceof datetime))
               {
                  // date values only
                  bdtValue = ((date) bdt).longValue();
               }
               if (bdt instanceof logical)
               {
                  // logical values
                  bdtValue = ((logical) bdt).booleanValue() ? 1L : 0L;
               }
               else if (bdt instanceof NumberType)
               {
                  // decimal, integer, int64, recid
                  bdtValue = ResourceIdHelper.idFromString(((NumberType) bdt).toStringMessage());
               }
               else if (bdt instanceof Text)
               {
                  // character or longchar values can be converted
                  bdtValue = ResourceIdHelper.idFromString(((Text) bdt).toStringMessage());
               }
            }
            finally
            {
               // nothings
            }   
         }
      }
      
      // If BDT value is null conversion is not possible.
      if (bdtValue == null)
      {
         incompatibleTypesOnConversion();
      }
      else
      {
         return resourceId.compareTo(bdtValue);
      }

      return -1;
   }
      
   /**
    * Creates a string representation of the instance data using the format
    * {@link ResourceIdHelper#DEFAULT_FORMAT}. If the instance represents the
    * <code>unknown value</code>, a '?' will be returned.
    *
    * @return   The formatted string.
    */
   public String toString()   
   {
      return toString(null);
   }
   
   /**
    * Creates a string representation of the instance data using the user
    * specified format string (Progress 4GL compatible) or if no such string
    * is provided, the default format of {@link ResourceIdHelper#DEFAULT_FORMAT}.
    * If the instance represents the <code>unknown value</code>, a '?' will be
    * returned.
    *
    * @param    fmt
    *           The Progress 4GL format string or <code>null</code> if the
    *           default format is to be used.
    *
    * @return   The formatted string.
    */
   public String toString(String fmt)   
   {
      if (isUnknown())
      {
         return ResourceIdHelper.UNKNOWN_STRING;
      }
      
      Long id = resourceId(value);
      return ResourceIdHelper.idToString(fmt, id);
   }
   
   /**
    * Creates a string representation of the instance data in a form that is
    * compatible with the <code>MESSAGE</code> language statement.  If the
    * instance represents the <code>unknown value</code>, a '?' will be
    * returned.
    *
    * @return   The 'message' formatted string.
    */
   public String toStringMessage()   
   {
      return toString();
   }
   
   /**
    * Creates a string representation of the instance data using the 'export'
    * format.  If the instance represents the <code>unknown value</code>, a
    * '?' will be returned.
    *
    * @return   The 'export' formatted string.
    */
   public String toStringExport()   
   {
      return toString();
   }   
   
   /**
    * Return the default display format string for this type.
    * <p>
    * Warning: this is larger than the Progress version because the range of
    * handle values can be a 64-bit integer in this implementation.
    *
    * @return   The default format string.
    */
   public String defaultFormatString()   
   {
      return ResourceIdHelper.DEFAULT_FORMAT;
   }      
      
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      throw new NotSerializableException("handle");
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      throw new NotSerializableException("handle");
   }
   
   /**
    * Getter for the wrapped resource.
    * 
    * @return   The {@link #value resource}. 
    */
   public WrappedResource getResource()
   {
      return value;
   }
   
   /**
    * Get resource ID as a long value.
    * 
    * @return  Resource ID if any or null.
    * 
    */
   public Long getResourceId()
   {
      return isUnknown() ? null : handle.resourceId(value);      
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {  
      /** 
       * Stores handles that have been converted to strings and may need
       * to be converted back into a handle.
       */
      private Map<Long, WrappedResource> idResources = new HashMap<Long, WrappedResource>();
      
      /**
       * Map of resources and their resource ids. Resource ids are tracked
       * here because they are naturally related to handles.
       */
      private Map<WrappedResource, Long> resourceIds = new HashMap<WrappedResource, Long>();
      
      private ArrayDeque<handle> invalidCalls = new ArrayDeque<handle>();
   }
   
   /**
    * Handler for the {@link #invalidAttrAccessProxy} proxy.
    */
   private static class InvalidAttributeAccess
   implements InvocationHandler
   {
      
      /**
       * This proxy is used only for cases when the handle is invalid or not of the expected type.
       */
      @Override
      public Object invoke(Object proxy, Method method, Object[] args)
      throws Throwable
      {
         WorkArea wa = work.get();
         
         BaseDataType unkn = null;
         Class<?> ret = method.getReturnType();

         if (ret == BaseDataType.class)
         {
            unkn = new unknown();
         }
         else if (ret != null && BaseDataType.class.isAssignableFrom(ret))
         {
            unkn = (BaseDataType) ret.newInstance();
         }

         handle target = wa.invalidCalls.pop();
         String attr = getAttrOrMethodName(method, args);

         if (!target._isValid())
         {
            // TODO: there are some cases (like FOCUS:attr or SELF:attr access, when FOCUS/SELF 
            // are unknown) which show only the 3140 error. this requires conversion changes, as
            // this is something related to the fact that the attr/mthd access is done directly
            // via a sys-handle, and not via an intermediate variable.
            
            final String msg1 =
               "Invalid handle.  Not initialized or points to a deleted object. (3135)";
            final int code1 = 3135;
            
            // special cases, depending on which method is executed
            String msg2 = "Cannot access the %s attribute because the widget does not exist. (3140)";
            msg2 = String.format(msg2, attr);
            int code2 = 3140; 

            final int[] errIds = { code1, code2 };
            final String[] msgs = { msg1, msg2 };
            
            ErrorManager.recordOrThrowError(errIds, msgs, false, true);

            return unkn;
         }

         // valid handle, wrong API, show error
         final String errMsg = 
            "**%s is not a queryable attribute for %s widget";
         String type = target.unwrapType().getResourceType().toStringMessage();
         String msg = String.format(errMsg, attr, type);
         ErrorManager.displayError(4052, msg, false);

         return unkn;
      }
   }
   
   /**
    * A "zero" pseudo-resource. A "zero" resource is used for handles created by
    * the {@link handle#fromString(String)} method when a formally valid ID is 
    * provided, but the ID doesn't denote a valid resource.
    * <p>
    * A zero resource is not valid but is not unknown.
    */
   private static final WrappedResource ZERO_RESOURCE = new WrappedResource()
   {
      
      @Override
      public boolean valid()
      {
         return false;
      }
      
      @Override
      public boolean unknown()
      {
         return false;
      }
   };
}
